# Copyright 2011 Fluidinfo Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

from twisted.trial import unittest
from twisted.internet import defer
from txrdq.dpq import _Task, DeferredPriorityQueue
import random


class TestTask(unittest.TestCase):
    def testAttrs(self):
        """
        Test that the basic attributes set by the _Task class
        actually work.
        """
        obj = 'x'
        priority = 10
        count = 5
        task = _Task(obj=obj, priority=priority, count=count)
        self.assertEqual(obj, task.obj)
        self.assertEqual(priority, task.priority)
        self.assertEqual(count, task.count)

    def testLessThan(self):
        """
        Test that a simple less-than comparison works.
        """
        t1 = _Task(obj='z', priority=5, count=0)
        t2 = _Task(obj='a', priority=10, count=0)
        self.assertTrue(t1 < t2)


class TestDeferredPriorityQueue(unittest.TestCase):
    def testSimplePriorities(self):
        """
        Objects added to the queue with different priorities should
        come off the queue in their correct priority order.
        """
        q = DeferredPriorityQueue()
        q.put('a', 20)
        q.put('b', 10)
        q.put('c', 15)
        gotten = []
        q.get().addCallback(gotten.append)
        q.get().addCallback(gotten.append)
        q.get().addCallback(gotten.append)
        self.assertEqual(['b', 'c', 'a'], gotten)

    def testRepeatedPut(self):
        """
        Add the same object to the queue twice and make sure
        C{RuntimeError} is raised.
        """
        q = DeferredPriorityQueue()
        q.put(self, 20)
        self.assertRaises(RuntimeError, q.put, self, 10)

    def testSimpleReprioritize(self):
        """
        Add 3 objects to the queue with different priorities, change one of
        the priorities and make sure they come out of the queue in their
        correct priority order.
        """
        q = DeferredPriorityQueue()
        q.put('a', 20)
        q.put('b', 10)
        q.put('c', 15)
        q.reprioritize('c', 7)
        q.reprioritize('a', 8)
        gotten = []
        q.get().addCallback(gotten.append)
        q.get().addCallback(gotten.append)
        q.get().addCallback(gotten.append)
        self.assertEqual(['c', 'a', 'b'], gotten)

    def testReprioritizeToSameValue(self):
        """
        Add 3 objects to the queue with different priorities, change one of
        the priorities to be the same as it was and make sure they come out
        of the queue in their correct priority order.
        """
        q = DeferredPriorityQueue()
        q.put('a', 1)
        q.put('b', 2)
        q.put('c', 3)
        q.reprioritize('b', 2)
        gotten = []
        q.get().addCallback(gotten.append)
        q.get().addCallback(gotten.append)
        q.get().addCallback(gotten.append)
        self.assertEqual(['a', 'b', 'c'], gotten)

    def testReprioritizeNonexistent(self):
        """
        Trying to reprioritize a non-existent task should raise a
        C{KeyError}.
        """
        q = DeferredPriorityQueue()
        self.assertRaises(KeyError, q.reprioritize, 'c', 7)

    def testReprioritizeDeleted(self):
        """
        Trying to reprioritize a non-existent task should raise a
        C{KeyError}.
        """
        q = DeferredPriorityQueue()
        q.put('a', 10)
        q.delete('a')
        self.assertRaises(KeyError, q.reprioritize, 'a', 7)

    def testDelete(self):
        """
        Delete an object from the queue, and make sure it doesn't show up
        in a get or somehow survive in the queue.
        """
        q = DeferredPriorityQueue()
        q.put('a', 0)
        q.delete('a')
        q.put('b', 0)
        gotten = []
        q.get().addCallback(gotten.append)
        self.assertEqual(['b'], gotten)
        self.assertEqual(len(q), 0)

    def testDeleteNonExistent(self):
        """
        Deleting a non-existent object should raise a C{KeyError}.
        """
        q = DeferredPriorityQueue()
        self.assertRaises(KeyError, q.delete, 'a')

    def testDeleteAlreadyDeleted(self):
        """
        Trying to delete an object that has been deleted should raise a
        C{KeyError}.
        """
        q = DeferredPriorityQueue()
        q.put('a', 0)
        q.delete('a')
        self.assertRaises(KeyError, q.delete, 'a')

    def testStable(self):
        """
        Objects added to the queue with identical priorities should come
        off the list in the order they entered.
        """
        n = 10
        q = DeferredPriorityQueue()
        values = range(n)
        random.shuffle(values)
        for value in values:
            q.put(value, 0)
        gotten = []
        for _ in range(n):
            q.get().addCallback(gotten.append)
        self.assertEqual(values, gotten)


class LegacyTwistedDeferredPriorityQueueTests(unittest.TestCase):
    """
    These are tests on C{DeferredQueue} from twisted/test/test_defer.py
    adapted to test the same things in C{DeferredPriorityQueue}.

    Priorities are always zero in the queue puts.
    """
    def testQueue(self):
        N, M = 2, 2
        queue = DeferredPriorityQueue(N, M)

        gotten = []

        for i in range(M):
            queue.get().addCallback(gotten.append)
        self.assertRaises(defer.QueueUnderflow, queue.get)

        for i in range(M):
            queue.put(obj=i, priority=0)
            self.assertEquals(gotten, range(i + 1))
        for i in range(N):
            queue.put(obj=N + i, priority=0)
            self.assertEquals(gotten, range(M))
        self.assertRaises(defer.QueueOverflow, queue.put, obj=None, priority=0)

        gotten = []
        for i in range(N):
            queue.get().addCallback(gotten.append)
            self.assertEquals(gotten, range(N, N + i + 1))

        queue = DeferredPriorityQueue()
        gotten = []
        for i in range(N):
            queue.get().addCallback(gotten.append)
        for i in range(N):
            queue.put(obj=i, priority=0)
        self.assertEquals(gotten, range(N))

        queue = DeferredPriorityQueue(size=0)
        self.assertRaises(defer.QueueOverflow, queue.put, obj=None, priority=0)

        queue = DeferredPriorityQueue(backlog=0)
        self.assertRaises(defer.QueueUnderflow, queue.get)

    def test_cancelQueueAfterSynchronousGet(self):
        """
        When canceling a L{Deferred} from a L{DeferredQueue} that already has
        a result, the cancel should have no effect.
        """
        def _failOnErrback(_):
            self.fail("Unexpected errback call!")

        queue = DeferredPriorityQueue()
        d = queue.get()
        d.addErrback(_failOnErrback)
        queue.put(obj=None, priority=0)
        d.cancel()

    def test_cancelQueueAfterGet(self):
        """
        When canceling a L{Deferred} from a L{DeferredQueue} that does not
        have a result (i.e., the L{Deferred} has not fired), the cancel
        causes a L{defer.CancelledError} failure. If the queue has a result
        later on, it doesn't try to fire the deferred.
        """
        def cb(ignore):
            # If the deferred is still linked with the deferred queue, it will
            # fail with an AlreadyCalledError
            queue.put(obj=None, priority=0)
            return queue.get().addCallback(self.assertIdentical, None)

        queue = DeferredPriorityQueue()
        d = queue.get()
        self.assertFailure(d, defer.CancelledError)
        d.cancel()
        return d.addCallback(cb)



# Copyright 2011 Fluidinfo Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

from twisted.trial import unittest
from txrdq.job import Job
from twisted.internet import reactor, task


class TestJob(unittest.TestCase):
    """
    Test the L{Job} class used by L{ResizableDispatchQueue}.

    Tests that involve cancellation come in almost-identical pairs. The
    first cancels the job by calling job.cancel() and the second cancels it
    by obtaining a watching C{Deferred} on the job and cancelling that.
    These two ways of cancelling a job will have an identical effect. The
    tests that cancel via cancelling the watching C{Deferred} have
    CancelDeferred in their names and follow immediately after the one that
    cancels the job directly.
    """

    def testInitiallyPending(self):
        """
        A new job should be in the PENDING state.
        """

        job = Job(None, 0)
        self.assertEqual(Job.PENDING, job.state)

    def testLaunchLaunchFails(self):
        """
        Launch a job twice and make sure the second attempt gets a
        C{RuntimeError}.
        """
        job = Job(None, 0)
        job.launch(lambda x: x)
        self.assertRaises(RuntimeError, job.launch, lambda x: x)

    def testWatchCancel(self):
        """
        Watch a job and check that when we cancel it before it has been
        launched the watcher gets an errback with a L{Job} instance whose
        state is CANCELLED.
        """

        def cb(result):
            self.fail('Unexpectedly succeeded!')

        def eb(failure):
            self.assertTrue(isinstance(failure.value, Job))
            self.assertEqual(failure.value.state, Job.CANCELLED)

        job = Job(None, 0)
        deferred = job.watch()
        deferred.addCallbacks(cb, eb)
        job.cancel()
        return deferred

    def testWatchCancelDeferred(self):
        """
        Watch a job and check that when we cancel the watching C{Deferred}
        before the job has been launched the watcher gets an errback with a
        L{Job} instance whose state is CANCELLED.
        """

        def cb(result):
            self.fail('Unexpectedly succeeded!')

        def eb(failure):
            self.assertTrue(isinstance(failure.value, Job))
            self.assertEqual(failure.value.state, Job.CANCELLED)

        job = Job(None, 0)
        deferred = job.watch()
        deferred.addCallbacks(cb, eb)
        deferred.cancel()
        return deferred

    def testWatchLaunchSucceedCheckResult(self):
        """
        Watch a job, launch it, and check the C{Deferred} receives a L{Job}
        instance whose result is correct.
        """

        def cb(result):
            self.assertEqual(result.result, 42)

        job = Job(6, 0)
        deferred = job.watch()
        deferred.addCallback(cb)
        job.launch(lambda x: x * 7)
        return deferred

    def testWatchLaunchSucceed(self):
        """
        Watch a job, launch it, and check the C{Deferred} receives a L{Job}
        instance whose state is FINISHED.
        """

        def cb(result):
            self.assertEqual(result.state, Job.FINISHED)

        job = Job(None, 0)
        deferred = job.watch()
        deferred.addCallback(cb)
        job.launch(lambda x: x)
        return deferred

    def testLaunchSucceedWatch(self):
        """
        Launch a job that succeeds, then watch it and check the C{Deferred}
        receives a L{Job} instance whose state is FINISHED.
        """

        def cb(result):
            self.assertEqual(result.state, Job.FINISHED)

        job = Job(None, 0)
        job.launch(lambda x: x)
        deferred = job.watch()
        deferred.addCallback(cb)
        return deferred

    def testWatchLaunchFail(self):
        """
        Watch a job that launches and fails. The L{Job} instance in the
        failure must be in the FAILED state. Check the underlying failure
        is as expected (in this case a ZeroDivisionError).
        """

        def cb(result):
            self.fail('Unexpectedly succeeded!')

        def eb(failure):
            self.assertEqual(failure.value.state, Job.FAILED)
            self.assertTrue(isinstance(failure.value.failure.value,
                                       ZeroDivisionError))

        job = Job(None, 0)
        deferred = job.watch()
        deferred.addCallbacks(cb, eb)
        job.launch(lambda x: 1 / 0)
        return deferred

    def testLaunchFailWatch(self):
        """
        Launch a job that fails, then watch it. The L{Job} instance
        received in the failure must be in the FAILED state. Check the
        underlying failure is as expected (in this case a
        ZeroDivisionError).
        """

        def cb(result):
            self.fail('Unexpectedly succeeded!')

        def eb(failure):
            self.assertEqual(failure.value.state, Job.FAILED)
            self.assertTrue(isinstance(failure.value.failure.value,
                                       ZeroDivisionError))

        job = Job(None, 0)
        job.launch(lambda x: 1 / 0)
        deferred = job.watch()
        deferred.addCallbacks(cb, eb)
        return deferred

    def testWatchLaunchSucceedCancel(self):
        """
        Watch a job that is launched and succeeds, and then call cancel on
        it.  The C{Deferred} will receive a Job instance whose state is
        FINISHED (i.e., the cancel will have no effect).
        """

        def cb(result):
            self.assertEqual(result.state, Job.FINISHED)

        job = Job(None, 0)
        deferred = job.watch()
        deferred.addCallback(cb)
        job.launch(lambda x: x)
        job.cancel()
        return deferred

    def testWatchLaunchSucceedCancelDeferred(self):
        """
        Watch a job that is launched and succeeds, and then call cancel on
        the watching C{Deferred}.  The C{Deferred} will receive a Job instance
        whose state is FINISHED (i.e., the cancel will have no effect).
        """

        def cb(result):
            self.assertEqual(result.state, Job.FINISHED)

        job = Job(None, 0)
        deferred = job.watch()
        deferred.addCallback(cb)
        job.launch(lambda x: x)
        deferred.cancel()
        return deferred

    def testWatchLaunchFailCancel(self):
        """
        Watch a job that is launched and fails, and then call cancel on
        it.  The C{Deferred} will receive a Job instance whose state is
        FAILED (i.e., the cancel will have no effect).
        """

        def cb(result):
            self.fail('Unexpectedly succeeded!')

        def eb(failure):
            self.assertEqual(failure.value.state, Job.FAILED)
            self.assertTrue(isinstance(failure.value.failure.value,
                                       ZeroDivisionError))

        job = Job(None, 0)
        deferred = job.watch()
        deferred.addCallbacks(cb, eb)
        job.launch(lambda x: 1 / 0)
        job.cancel()
        return deferred

    def testWatchLaunchFailCancelDeferred(self):
        """
        Watch a job that is launched and fails, and then call cancel on the
        watching C{Deferred}.  The C{Deferred} will receive a Job instance
        whose state is FAILED (i.e., the cancel will have no effect).
        """

        def cb(result):
            self.fail('Unexpectedly succeeded!')

        def eb(failure):
            self.assertEqual(failure.value.state, Job.FAILED)
            self.assertTrue(isinstance(failure.value.failure.value,
                                       ZeroDivisionError))

        job = Job(None, 0)
        deferred = job.watch()
        deferred.addCallbacks(cb, eb)
        job.launch(lambda x: 1 / 0)
        deferred.cancel()
        return deferred

    def testLaunchFailWatchCancel(self):
        """
        Launch a job that fails, then watch it, then cancel it.  The
        C{Deferred} will receive a Job instance whose state is FAILED (i.e.,
        the cancel will have no effect).
        """

        def cb(result):
            self.fail('Unexpectedly succeeded!')

        def eb(failure):
            self.assertEqual(failure.value.state, Job.FAILED)
            self.assertTrue(isinstance(failure.value.failure.value,
                                       ZeroDivisionError))

        job = Job(None, 0)
        job.launch(lambda x: 1 / 0)
        deferred = job.watch()
        deferred.addCallbacks(cb, eb)
        job.cancel()
        return deferred

    def testLaunchFailWatchCancelDeferred(self):
        """
        Launch a job that fails, then watch it, then cancel the watching
        C{Deferred}.  The C{Deferred} will receive a Job instance whose
        state is FAILED (i.e., the cancel will have no effect).
        """

        def cb(result):
            self.fail('Unexpectedly succeeded!')

        def eb(failure):
            self.assertEqual(failure.value.state, Job.FAILED)
            self.assertTrue(isinstance(failure.value.failure.value,
                                       ZeroDivisionError))

        job = Job(None, 0)
        job.launch(lambda x: 1 / 0)
        deferred = job.watch()
        deferred.addCallbacks(cb, eb)
        deferred.cancel()
        return deferred

    def testWatchLaunchUnderwayCancel(self):
        """
        Watch a job, launch it with a function that returns a C{Deferred}
        that is slow to fire, then cancel the job. Check the watching
        C{Deferred} receives a L{Job} instance whose state is CANCELLED.
        """

        def cb(result):
            self.fail('Unexpectedly succeeded!')

        def eb(failure):
            self.assertEqual(failure.value.state, Job.CANCELLED)

        def run(x):
            return task.deferLater(reactor, 3, lambda x: x)

        job = Job(None, 0)
        deferred = job.watch()
        deferred.addCallbacks(cb, eb)
        job.launch(run)
        job.cancel()
        return deferred

    def testWatchLaunchUnderwayCancelDeferred(self):
        """
        Watch a job, launch it with a function that returns a C{Deferred}
        that is slow to fire, then cancel the watching C{Deferred}. Check
        the watching C{Deferred} receives a L{Job} instance whose state is
        CANCELLED.
        """

        def cb(result):
            self.fail('Unexpectedly succeeded!')

        def eb(failure):
            self.assertEqual(failure.value.state, Job.CANCELLED)

        def run(x):
            return task.deferLater(reactor, 3, lambda x: x)

        job = Job(None, 0)
        deferred = job.watch()
        deferred.addCallbacks(cb, eb)
        job.launch(run)
        deferred.cancel()
        return deferred



# Copyright 2010, 2011 Fluidinfo Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you
# may not use this file except in compliance with the License.  You
# may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.  See the License for the specific language governing
# permissions and limitations under the License.

from twisted.internet import defer


class DeferredPool(object):
    """
    Maintain a pool of not-yet-fired deferreds and provide a mechanism to
    request a deferred that fires when the pool size goes to zero.
    """
    def __init__(self):
        self._pool = set()
        self._waiting = []

    def _fired(self, result, d):
        """
        Callback/errback each pooled deferred runs when it fires. The
        deferred first removes itself from the pool. If the pool is then
        empty, fire all the waiting deferreds (which were returned by
        notifyWhenEmpty).
        """
        self._pool.remove(d)
        if not self._pool:
            waiting, self._waiting = self._waiting, []
            for waiter in waiting:
                waiter.callback(None)
        return result

    def add(self, d):
        """
        Add a deferred to the pool.
        """
        d.addBoth(self._fired, d)
        self._pool.add(d)
        return d

    def notifyWhenEmpty(self, testImmediately=True):
        """
        Return a deferred that fires (with None) when the pool empties.  If
        testImmediately is True and the pool is empty, return an already
        fired deferred (via succeed).
        """
        if testImmediately and not self._pool:
            return defer.succeed(None)
        else:
            d = defer.Deferred()
            self._waiting.append(d)
            return d



#!/usr/bin/env python

import os


# Utility function to read the README file.
# Credit: http://pypi.python.org/pypi/an_example_pypi_project
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

d = dict(name='txRDQ',
         version='0.2.8',
         provides=['txrdq'],
         maintainer='Fluidinfo Inc.',
         maintainer_email='info@fluidinfo.com',
         url='https://launchpad.net/txrdq',
         download_url='https://code.launchpad.net/txrdq',
         packages=['txrdq', 'txrdq.test'],
         keywords=['twisted dispatch job queue'],
         classifiers=[
             'Programming Language :: Python',
             'Framework :: Twisted',
             'Development Status :: 4 - Beta',
             'Intended Audience :: Developers',
             'License :: OSI Approved :: Apache Software License',
             'Operating System :: OS Independent',
             'Topic :: Software Development :: Libraries :: Python Modules',
             ],
         description=('A Twisted class for queueing and dispatching jobs in '
                      'a controlled manner.'),
         long_description=read('README'),
         )

try:
    from setuptools import setup
    _ = setup  # Keeps pyflakes from complaining.
except ImportError:
    from distutils.core import setup
else:
    d['install_requires'] = ['Twisted']

setup(**d)
